// Arm controller v3
//
// Changelog:
//  2015-05-19:
//    v3 - compiled with a new TaskScheduler version 1.4 (sleep IDLE instead of delay)
//    v3 - buttons moved to pin state change interrupts instead of polling

// INCLUDES

#include <DirectIO.h>

#include <SPI.h>
#include "nRF24L01.h"
#include "RF24.h"

#include <TaskScheduler.h>
#include <PinChangeInt.h>

// TEST AND DEBUG DEFINES 

//#define _DEBUG_
//#define _TEST_


// DEFINES

#define  CE_PIN   8
#define  CSN_PIN  7

#define  PWM_M1_MIN  120
#define  PWM_M1_MAX  200
#define  PWM_M1_INC  20

#define  PWM_M2_MIN  120
#define  PWM_M2_MAX  160
#define  PWM_M2_INC  20

#define  PWM_M3_MIN  120
#define  PWM_M3_MAX  180
#define  PWM_M3_INC  40

#define  PWM_M4_MIN  120
#define  PWM_M4_MAX  180
#define  PWM_M4_INC  40

#define  PWM_M5_MIN  120
#define  PWM_M5_MAX  200
#define  PWM_M5_INC  40

#define  GO_PIN      A3
#define  STOP_PIN  A2

Output<A0>            pLed;
Output<A1>            pSS1;
Input<STOP_PIN>    pStopButton;
Input<GO_PIN>       pGoButton;
Output<A4>            pBaseRight;
Output<A5>            pCommsLed;
AnalogOutput<3>     pBasePwm=0;
Output<4>               pBaseLeft;
AnalogOutput<5>     pArmPwm=0;
AnalogOutput<6>     pElbowPwm=0;
AnalogOutput<9>     pWristPwm=0;
AnalogOutput<10>   pClawPwm=0;


#define M_OFF     0b00
#define M_ON      0b11
#define M_RIGHT   0b01
#define M_LEFT    0b10

#define  RADIO_DEFAULT_CHANNEL  56

// STRUCTURES

struct packed_bits {
  byte m1  : 2; // claw
  byte m2  : 2; // wrist
  byte m3  : 2; // elbow
  byte m4  : 2; // arm
  byte m5  : 2; // base rotate
  byte led : 2;
} ctrlWord, cwRx, cwPrev, cwActive;

struct {
  int m1;
  int m2;
  int m3;
  int m4;
  int m5;
} pwms;

// TASK DEFINES

#define RADIO_PERIOD    500   // radio -event driven via interrupts

#ifdef _DEBUG_
#define RADIO_TIMEOUT   2000   // 2 seconds
#else
#define RADIO_TIMEOUT   10000   // 2 seconds
#endif

#define CTRL_PERIOD      1000   // 1 times/sec
#define MCTRL_PERIOD    1000   // normally speed increases in appropriate steps every N second(s)
#define MCTRL_REV_PER     50   // for one-time reverse we need to give full speed reverse for N ms
//#define BUTTON_PERIOD   100 


// TASKS

Scheduler runner;

Task tRadio (RADIO_PERIOD, -1, &radioCallback);
Task tRadioTimeout (RADIO_TIMEOUT, -1, &radioTimeoutCallback);
Task tArmControl (CTRL_PERIOD, -1, &armControlCallback);
Task tM1Control (MCTRL_PERIOD, -1, &m1SpeedControl);
Task tM2Control (MCTRL_PERIOD, -1, &m2SpeedControl);
Task tM3Control (MCTRL_PERIOD, -1, &m3SpeedControl);
Task tM4Control (MCTRL_PERIOD, -1, &m4SpeedControl);
Task tM5Control (MCTRL_PERIOD, -1, &m5SpeedControl);
Task tGreenButton (0, 1, &greenButtonCallback);
Task tRedButton (0, 1, &redButtonCallback);

#ifdef _DEBUG_
Task tDisplay (1000, -1, &displayCallback);
#endif


// VARIABLES

const uint64_t pipe = 0xE8E8F0F0E1LL; // Define the transmit pipe
volatile bool dsr;  // data set ready

// DRIVERS

RF24 radio(CE_PIN, CSN_PIN);


// CODE: CALLBACKS

void radioCallback() {
#ifdef _DEBUG_
//  Serial.print("radioCallback: dsr="); 
//  Serial.print(dsr); 
//  Serial.print(", radio.avail?=");
//  Serial.print(radio.available());
//  Serial.println();
#endif

  pCommsLed = LOW;
  if (dsr) {
      noInterrupts();
      ctrlWord = cwRx;
      dsr = false;
      interrupts();
      tArmControl.enable();
  }
}


void radioTimeoutCallback() {
#ifdef _DEBUG_
  Serial.println("Radio timeout!");
#endif
  stopMotors();
  halt();
}


void armControlCallback() {
  
  bool changed = false;

#ifdef _DEBUG_
//  Serial.print("armControlCallback"); 
//  Serial.println();
#endif
  
// Only do anything if changes were requested  
// LED - just execute the change immediately

// For motors - two different approaches:
// 1. If motor was stopped - start execution in the requested direction with gradual increase of speed
// 2. If the motor was moving and is requested to stop - initiate a momentary reverse pulse to break
// 3. If the motor was moving in one direction and is immediately requested to reverse do #2, then #1

  if ( !cwEqual(cwPrev, ctrlWord) ) {
    
    if (cwPrev.led != ctrlWord.led) {
      cwActive.led = ctrlWord.led;
      changed = true;
    }
    
    if (cwPrev.m1 != ctrlWord.m1) {
      if (cwPrev.m1 == M_OFF) {
        cwActive.m1 = ctrlWord.m1;
        pwms.m1 = PWM_M1_MIN;
        tM1Control.set(MCTRL_PERIOD, -1, &m1SpeedControl);
        tM1Control.enableDelayed();
        changed = true;
      }
      if (ctrlWord.m1 == M_OFF) {
        cwActive.m1 = ~cwPrev.m1;  // reverse direction
        pwms.m1 = PWM_M1_MAX;
        tM1Control.set(MCTRL_REV_PER, 1, &m1SpeedControlRevStop);
        tM1Control.enableDelayed();
        changed = true;
      }
       if ( (cwPrev.m1 | ctrlWord.m1) == 0b11) {
        cwActive.m1 = ~cwPrev.m1;  // reverse direction
        pwms.m1 = PWM_M1_MAX;
        tM1Control.set(MCTRL_REV_PER, -1, &m1SpeedControlRevGrad);
        tM1Control.enableDelayed();
        changed = true;
      }
    }
   
    if (cwPrev.m2 != ctrlWord.m2) {
      if (cwPrev.m2 == M_OFF) {
        cwActive.m2 = ctrlWord.m2;
        pwms.m2 = PWM_M2_MIN;
        tM2Control.set(MCTRL_PERIOD, -1, &m2SpeedControl);
        tM2Control.enableDelayed();
        changed = true;
      }
      if (ctrlWord.m2 == M_OFF) {
        cwActive.m2 = ~cwPrev.m2;  // reverse direction
        pwms.m2 = PWM_M2_MAX;
        tM2Control.set(MCTRL_REV_PER, 1, &m2SpeedControlRevStop);
        tM2Control.enableDelayed();
        changed = true;
      }
       if ( (cwPrev.m2 | ctrlWord.m2) == 0b11) {
        cwActive.m2 = ~cwPrev.m2;  // reverse direction
        pwms.m2 = PWM_M2_MAX;
        tM2Control.set(MCTRL_REV_PER, -1, &m2SpeedControlRevGrad);
        tM2Control.enableDelayed();
        changed = true;
      }
    }
   
    if (cwPrev.m3 != ctrlWord.m3) {
      if (cwPrev.m3 == M_OFF) {
        cwActive.m3 = ctrlWord.m3;
        pwms.m3 = PWM_M3_MIN;
        tM3Control.set(MCTRL_PERIOD, -1, &m3SpeedControl);
        tM3Control.enableDelayed();
        changed = true;
      }
      if (ctrlWord.m3 == M_OFF) {
        cwActive.m3 = ~cwPrev.m3;  // reverse direction
        pwms.m3 = PWM_M3_MAX;
        tM3Control.set(MCTRL_REV_PER, 1, &m3SpeedControlRevStop);
        tM3Control.enableDelayed();
        changed = true;
      }
       if ( (cwPrev.m3 | ctrlWord.m3) == 0b11) {
        cwActive.m3 = ~cwPrev.m3;  // reverse direction
        pwms.m3 = PWM_M3_MAX;
        tM3Control.set(MCTRL_REV_PER, -1, &m3SpeedControlRevGrad);
        tM3Control.enableDelayed();
        changed = true;
      }
    }
   
    if (cwPrev.m4 != ctrlWord.m4) {
      if (cwPrev.m4 == M_OFF) {
        cwActive.m4 = ctrlWord.m4;
        pwms.m4 = PWM_M4_MIN;
        tM4Control.set(MCTRL_PERIOD, -1, &m4SpeedControl);
        tM4Control.enableDelayed();
        changed = true;
      }
      if (ctrlWord.m4 == M_OFF) {
        cwActive.m4 = ~cwPrev.m4;  // reverse direction
        pwms.m4 = PWM_M4_MAX;
        tM4Control.set(MCTRL_REV_PER, 1, &m4SpeedControlRevStop);
        tM4Control.enableDelayed();
        changed = true;
      }
       if ( (cwPrev.m4 | ctrlWord.m4) == 0b11) {
        cwActive.m4 = ~cwPrev.m4;  // reverse direction
        pwms.m4 = PWM_M4_MAX;
        tM4Control.set(MCTRL_REV_PER, -1, &m4SpeedControlRevGrad);
        tM4Control.enableDelayed();
        changed = true;
      }
    }

    if (cwPrev.m5 != ctrlWord.m5) {
      if (cwPrev.m5 == M_OFF) {
        cwActive.m5 = ctrlWord.m5;
        pwms.m5 = PWM_M5_MIN;
        tM5Control.set(MCTRL_PERIOD, -1, &m5SpeedControl);
        tM5Control.enableDelayed();
        changed = true;
      }
      if (ctrlWord.m5 == M_OFF) {
        cwActive.m5 = ~cwPrev.m5;  // reverse direction
        pwms.m5 = PWM_M5_MAX;
        tM5Control.set(MCTRL_REV_PER, 1, &m5SpeedControlRevStop);
        tM5Control.enableDelayed();
        changed = true;
      }
       if ( (cwPrev.m5 | ctrlWord.m5) == 0b11) {
        cwActive.m5 = ~cwPrev.m5;  // reverse direction
        pwms.m5 = PWM_M5_MAX;
        tM5Control.set(MCTRL_REV_PER, -1, &m5SpeedControlRevGrad);
        tM5Control.enableDelayed();
        changed = true;
      }
    }

    cwPrev = ctrlWord;
    if (changed) controlMotors();
  }
}

void controlMotors() {
  
  byte *p = (byte *) &cwActive;
 
  if (cwActive.led == M_ON) pLed = HIGH;
  else pLed = LOW;

  if (cwActive.m1 == M_OFF) pwms.m1 = 0;
  if (cwActive.m2 == M_OFF) pwms.m2 = 0;
  if (cwActive.m3 == M_OFF) pwms.m3 = 0;
  if (cwActive.m4 == M_OFF) pwms.m4 = 0;
  if (cwActive.m5 == M_OFF) pwms.m5 = 0;
  else {
    pBaseLeft = (cwActive.m5 == M_LEFT);
    pBaseRight = (cwActive.m5 == M_RIGHT);
  }  
  
  pClawPwm = pwms.m1;
  pWristPwm = pwms.m2;
  pElbowPwm = pwms.m3;
  pArmPwm = pwms.m4;
  pBasePwm = pwms.m5;
  
// Shift out pin assignments to the shift register via SPI
  csn(LOW);
  SPI.transfer(*p);
  csn(HIGH);
}

void csn(int mode)
{
  // Minimum ideal SPI bus speed is 2x data rate
  // If we assume 2Mbs data rate and 16Mhz clock, a
  // divider of 4 is the minimum we want.
  // CLK:BUS 8Mhz:2Mhz, 16Mhz:4Mhz, or 20Mhz:5Mhz
  if (mode == LOW) {
    SPI.setBitOrder(MSBFIRST);
    SPI.setDataMode(SPI_MODE0);
    SPI.setClockDivider(SPI_CLOCK_DIV4);
  }
  pSS1 = mode; 
}



void m1SpeedControl() {
  pwms.m1 += PWM_M1_INC;
  if (pwms.m1 >= PWM_M1_MAX) {
    pwms.m1 = PWM_M1_MAX;
    tM1Control.disable();
  }
  controlMotors();
}
void m1SpeedControlRevStop() {
  cwActive.m1 = M_OFF;
  pwms.m1 = 0;
  controlMotors();
}
void m1SpeedControlRevGrad() {
  cwActive.m1 = ctrlWord.m1;
  pwms.m1 = PWM_M1_MIN;
  tM1Control.set(MCTRL_PERIOD, -1, &m1SpeedControl);
  tM1Control.enableDelayed();
  controlMotors();
}


void m2SpeedControl() {
  pwms.m2 += PWM_M2_INC;
  if (pwms.m2 >= PWM_M2_MAX) {
    pwms.m2 = PWM_M2_MAX;
    tM2Control.disable();
  }
  controlMotors();
}
void m2SpeedControlRevStop() {
  cwActive.m2 = M_OFF;
  pwms.m2 = 0;
  controlMotors();
}
void m2SpeedControlRevGrad() {
  cwActive.m2 = ctrlWord.m2;
  pwms.m2 = PWM_M2_MIN;
  tM2Control.set(MCTRL_PERIOD, -1, &m2SpeedControl);
  tM2Control.enableDelayed();
  controlMotors();
}



void m3SpeedControl() {
  pwms.m3 += PWM_M3_INC;
  if (pwms.m3 >= PWM_M3_MAX) {
    pwms.m3 = PWM_M3_MAX;
    tM3Control.disable();
  }
  controlMotors();
}
void m3SpeedControlRevStop() {
  cwActive.m3 = M_OFF;
  pwms.m3 = 0;
  controlMotors();
}
void m3SpeedControlRevGrad() {
  cwActive.m3 = ctrlWord.m3;
  pwms.m3 = PWM_M3_MIN;
  tM3Control.set(MCTRL_PERIOD, -1, &m3SpeedControl);
  tM3Control.enableDelayed();
  controlMotors();
}



void m4SpeedControl() {
  pwms.m4 += PWM_M4_INC;
  if (pwms.m4 >= PWM_M4_MAX) {
    pwms.m4 = PWM_M4_MAX;
    tM4Control.disable();
  }
  controlMotors();
}
void m4SpeedControlRevStop() {
  cwActive.m4 = M_OFF;
  pwms.m4 = 0;
  controlMotors();
}
void m4SpeedControlRevGrad() {
  cwActive.m4 = ctrlWord.m4;
  pwms.m4 = PWM_M4_MIN;
  tM4Control.set(MCTRL_PERIOD, -1, &m4SpeedControl);
  tM4Control.enableDelayed();
  controlMotors();
}



void m5SpeedControl() {
  pwms.m5 += PWM_M5_INC;
  if (pwms.m5 >= PWM_M5_MAX) {
    pwms.m5 = PWM_M5_MAX;
    tM5Control.disable();
  }
  controlMotors();
}
void m5SpeedControlRevStop() {
  cwActive.m5 = M_OFF;
  pwms.m5 = 0;
  controlMotors();
}
void m5SpeedControlRevGrad() {
  cwActive.m5 = ctrlWord.m5;
  pwms.m5 = PWM_M5_MIN;
  tM5Control.set(MCTRL_PERIOD, -1, &m5SpeedControl);
  tM5Control.enableDelayed();
  controlMotors();
}



//#define  GO_PIN      A3
//#define  STOP_PIN  A2

void initButtons() {
  int buttonPressed=PCintPort::arduinoPin;
  if (buttonPressed == GO_PIN) tGreenButton.restart();
  if (buttonPressed == STOP_PIN) tRedButton.restart();
}

void redButtonCallback() {
#ifdef _DEBUG_
  Serial.println("redButtonCallback");
#endif
    halt();
}

void greenButtonCallback() {
#ifdef _DEBUG_
  Serial.println("greenButtonCallback");
#endif
    restart();
}


#ifdef _DEBUG_
void displayCallback() {
  char  ln[256];
  byte *p;
  
  p = (byte *) &ctrlWord;
  
  sprintf(ln, "led=%02d\tm1=%02d\tm2=%02d\tm3=%02d\tm4=%02d\tm5=%02d",ctrlWord.led,ctrlWord.m1,ctrlWord.m2,ctrlWord.m3,ctrlWord.m4,ctrlWord.m5);  
  Serial.println(ln);
  Serial.print(*p, HEX); p++; Serial.print(":"); Serial.println(*p, HEX);
}
#endif


// CODE: MAIN

void initPins() {
  pSS1 = HIGH;
  pinMode(2, INPUT);
  pinMode(3, OUTPUT); 
  pinMode(5, OUTPUT); 
  pinMode(6, OUTPUT); 
  pinMode(9, OUTPUT); 
  pinMode(10, OUTPUT);  // they say this is needed for SPI to be a master
//  pinMode(A4, OUTPUT);
}


void stopMotors() {

  initCtrlWord();
  cwRx = ctrlWord;
  cwPrev = ctrlWord;
  cwActive = ctrlWord;

  pwms.m1 = 0;
  pwms.m2 = 0;
  pwms.m3 = 0;
  pwms.m4 = 0;
  pwms.m5 = 0;
  
  controlMotors();
}

void initCtrlWord() {
  ctrlWord.led = M_OFF;
  ctrlWord.m1 = M_OFF;
  ctrlWord.m2 = M_OFF;
  ctrlWord.m3 = M_OFF;
  ctrlWord.m4 = M_OFF;
  ctrlWord.m5 = M_OFF;
}

void initRadio() {

//  radio.setChannel(RADIO_DEFAULT_CHANNEL);  // 0-127
  radio.setDataRate(RF24_250KBPS);  // RF24_1MBPS = 0, RF24_2MBPS, RF24_250KBPS
//  radio.setRetries(5,5);
  radio.setPayloadSize( sizeof(ctrlWord) );
//  radio.setAutoAck(true);
//  radio.enableAckPayload();
  radio.setCRCLength(RF24_CRC_8);  // RF24_CRC_DISABLED = 0, RF24_CRC_8, RF24_CRC_16
//  radio.setPALevel(RF24_PA_HIGH); // RF24_PA_MIN=-18dBm, RF24_PA_LOW=-12dBm, RF24_PA_MED=-6dBM, and RF24_PA_HIGH=0dBm. *  RF24_PA_MIN = 0,RF24_PA_LOW, RF24_PA_HIGH, RF24_PA_MAX, RF24_PA_ERROR
//  radio.testCarrier();
  delay(10);
}


bool cwEqual(struct packed_bits& a, struct packed_bits& b) {
  return ( (a.led == b.led) && (a.m1 == b.m1) && (a.m2 == b.m2) && (a.m3 == b.m3) && (a.m4 == b.m4) && (a.m5 == b.m5) );
}


void restart() {
#ifdef _DEBUG_
  Serial.println("restart");
#endif  
  
  runner.disableAll();
  tRadio.enable();
  tRadioTimeout.enableDelayed();
  tArmControl.enable();
  tRedButton.enable();

  radio.startListening();

  PCintPort::detachInterrupt(GO_PIN);
  PCintPort::attachInterrupt(STOP_PIN, &initButtons, RISING); 
  
#ifdef _DEBUG_
  runner.addTask(tDisplay);
  tDisplay.enable();
#endif
}


void halt() {
#ifdef _DEBUG_
  Serial.println("halt");
#endif  

  stopMotors();
  runner.disableAll();
  tGreenButton.enable();
  radio.powerDown();
  pCommsLed = LOW;

  PCintPort::detachInterrupt(STOP_PIN);
  PCintPort::attachInterrupt(GO_PIN, &initButtons, RISING); 
}


void setup() {
  initPins();
  
  SPI.begin();
  stopMotors();
  
#ifdef _DEBUG_
  Serial.begin(115200);
  Serial.println("Robotic Arm Motor Controller Circuit test");
#endif

#ifndef _DEBUG_
  power_adc_disable();
  power_twi_disable();
  power_usart0_disable();
#endif

  radio.begin();
  initRadio();
  radio.openReadingPipe(1, pipe);
  
  dsr = false;
  attachInterrupt(0, check_radio, FALLING);
  
  runner.init();
  
  runner.addTask(tRadio);
  runner.addTask(tRadioTimeout);
  runner.addTask(tArmControl);
  runner.addTask(tM1Control);
  runner.addTask(tM2Control);
  runner.addTask(tM3Control);
  runner.addTask(tM4Control);
  runner.addTask(tM5Control);
  runner.addTask(tGreenButton);
  runner.addTask(tRedButton);

  halt();
}


void loop() {
  runner.execute();
}


void check_radio(void)
{
  // What happened?
  bool tx,fail,rx;
  radio.whatHappened(tx,fail,rx);

  if ( rx )
  {
    dsr = true;
    radio.read( (byte*) &cwRx, sizeof(cwRx) );
    pCommsLed = HIGH;
    tRadioTimeout.delay();
    tRadio.enable();
  }
}
